# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Key ACLS Service (KACLS) API."""

import hashlib
import hmac
from drive_cse_upload import _cse_util
import requests


class CseKaclsClient(object):
  """KACLS API Client."""

  PRIV_WRAP = '/privilegedwrap'
  PRIV_UNWRAP = '/privilegedunwrap'

  def __init__(self):
    pass

  def privileged_wrap(
      self, key, resource_name, authn, kacls_url, perimeter_id=''
  ):
    """Wrap a data encryption key by a KACLS.

    Args:
        key: The base64-encoded data encryption key to wrap.
        resource_name: Name of the encrypted resource (Drive document).
        authn: An authentication token for the calling user.
        kacls_url: Base URL of the KACLS to use.
        perimeter_id: Optional perimeter id.

    Returns:
        The base64-encoded wrapped key computed by the KACLS.

    Raises:
      RuntimeError: An error occurred while calling the KACLS.
    """
    wrap_url = kacls_url + self.PRIV_WRAP
    request = {
        'key': key,
        'resource_name': resource_name,
        'perimeter_id': perimeter_id,
        'authentication': authn,
        'reason': 'import',
    }
    r = requests.post(wrap_url, json=request)
    if r.status_code != requests.codes.ok:
      raise RuntimeError('Wrap failed: ' + r.text)
    response = r.json()
    return response['wrapped_key']

  def privileged_unwrap(self, wrapped_key, resource_name, authn, kacls_url):
    """Unwrap a data encryption key by a KACLS.

    Args:
        wrapped_key: The base64-encoded wrapped key to unwrap.
        resource_name: Name of the encrypted resource (Drive document).
        authn: An authentication token for the calling user.
        kacls_url: Base URL of the KACLS to use.

    Returns:
        The base64-encoded unwrapped key computed by the KACLS.

    Raises:
      RuntimeError: An error occurred while calling the KACLS.
    """
    unwrap_url = kacls_url + self.PRIV_UNWRAP
    request = {
        'wrapped_key': wrapped_key,
        'resource_name': resource_name,
        'authentication': authn,
        'reason': 'import',
    }
    r = requests.post(unwrap_url, json=request)
    if r.status_code != requests.codes.ok:
      raise RuntimeError('Unwrap failed: ' + r.text)
    response = r.json()
    return response['key']

  def resource_key_hash(self, key, resource_name, perimeter_id=''):
    """Compute resource-key-hash.

    Args:
        key: The data encryption key.
        resource_name: Name of the encrypted resource (Drive document).
        perimeter_id: Optional perimeter id.

    Returns:
        A locally-computed base64-encoded resource_key_hash.
    """
    data = f'ResourceKeyDigest:{resource_name}:{perimeter_id}'.encode()
    digest = hmac.digest(key, data, hashlib.sha256)
    return _cse_util.b64encode(digest)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example for CSE Upload."""

import argparse
import pathlib
import sys
from drive_cse_upload import cse_uploader


def main():
  parser = argparse.ArgumentParser(
      prog='example', usage='%(prog)s [options...] <input-file>'
  )
  parser.add_argument(
      '--sa-key-file',
      dest='sa_key_file',
      required=True,
      help='service-account private key file [required]',
  )
  parser.add_argument(
      '--client-secret-file',
      dest='client_secret_file',
      required=True,
      help='oauth client secret file [required]',
  )
  parser.add_argument(
      '--saved-creds-file',
      dest='saved_creds_file',
      required=True,
      help='saved credentials file [required]',
  )
  parser.add_argument(
      '--as-user',
      dest='as_user',
      required=True,
      help='upload the file as this user [required]',
  )
  parser.add_argument(
      '--parent-id',
      dest='parent_id',
      help='parent folder/shared-drive for the uploaded file [None]',
  )
  parser.add_argument(
      '--validate',
      dest='validate',
      action=argparse.BooleanOptionalAction,
      default=True,
      help='validate the upload [True]',
  )
  parser.add_argument('input_file', metavar='input-file', help='file to upload')
  args = parser.parse_args()

  uploader = cse_uploader.CseUploader(
      args.sa_key_file,
      args.client_secret_file,
      args.saved_creds_file,
      args.validate,
  )
  filename = pathlib.Path(args.input_file).name
  try:
    file = uploader.upload(args.input_file, args.as_user, args.parent_id)
    drive_id = file['id']
    print(f'Uploaded {filename} as {drive_id}')
  except RuntimeError as e:
    print(f'File {filename} could not be uploaded: {e}')
    sys.exit(1)

if __name__ == '__main__':
  main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from drive_cse_upload import _cse_util
from absl.testing import absltest


class CseUtilTest(absltest.TestCase):

  def test_b64encode(self):
    data = b'i\x9f\xbb\xeb\xfc'
    self.assertEqual(_cse_util.b64encode(data), 'aZ+76/w=')

  def test_b64urlencode(self):
    data = b'i\x9f\xbb\xeb\xfc'
    self.assertEqual(_cse_util.b64urlencode(data), 'aZ-76_w=')

  def test_b64decode(self):
    data = 'aZ+76/w='
    self.assertEqual(_cse_util.b64decode(data), b'i\x9f\xbb\xeb\xfc')

  def test_b64urldecode(self):
    data = 'aZ-76_w='
    self.assertEqual(_cse_util.b64urldecode(data), b'i\x9f\xbb\xeb\xfc')

  def test_b64_roundtrip(self):
    data = random.randbytes(random.randint(2, 257))
    self.assertEqual(data, _cse_util.b64decode(_cse_util.b64encode(data)))

  def test_b64url_roundtrip(self):
    data = random.randbytes(random.randint(2, 257))
    self.assertEqual(data, _cse_util.b64urldecode(_cse_util.b64urlencode(data)))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Identity Provider (IDP) API."""

import datetime
import json
import os
import threading
from google.auth.exceptions import RefreshError
from google.auth.transport.requests import Request
from google.oauth2 import credentials
from google_auth_oauthlib.flow import InstalledAppFlow


class CseIdpClient(object):
  """Identity Provider (IDP) Client."""

  SCOPES = ['openid', 'email']

  def __init__(self, client_secret_file, saved_creds_file):
    """Constructor.

    Note: All the arguments denote files containing sensitive information that
    should be protected. Users must ensure that the files passed-in / created
    here are not readable by anyone but their owner.

    Args:
        client_secret_file: Client Secret File path for the configured IDP OAuth
          Client-Id; callers must download the file from their IDP and store it
          to this path.
        saved_creds_file: Where to store the IDP Oauth credentials, created by
          this module.
    """
    self._client_secret_file = client_secret_file
    self._saved_creds_file = saved_creds_file
    self._creds: credentials.Credentials = None
    self._lock = threading.Lock()

  def get_authn_token(self):
    """Get an OAuth id-token.

    Returns:
      An OAuth id-token for the calling user.
    """
    with self._lock:
      self._get_creds()
      return self._creds.id_token

  def _get_creds(self):
    """Get credentials."""

    # Google Python client library work-around for servers changing scopes
    os.environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = 'True'

    if not self._creds:
      self._creds = self._read_creds_from_file()

    if self._creds and self._creds.valid:
      return

    if self._creds:
      try:
        self._creds.refresh(Request())
        if self._creds.valid:
          return
      except RefreshError:
        pass

    flow = InstalledAppFlow.from_client_secrets_file(
        self._client_secret_file, scopes=self.SCOPES
    )
    self._creds = flow.run_local_server(open_browser=False, port=0)
    self._write_creds_to_file()

  def _read_creds_from_file(self):
    """Read credentials from file."""
    creds = None
    if os.path.exists(self._saved_creds_file):
      try:
        with open(self._saved_creds_file, 'r') as f:
          creds_json = json.load(f)
        creds = credentials.Credentials(
            token=creds_json.get('access_token'),
            refresh_token=creds_json.get('refresh_token'),
            id_token=creds_json.get('id_token'),
            token_uri=creds_json.get('token_uri'),
            client_id=creds_json.get('client_id'),
            client_secret=creds_json.get('client_secret'),
            scopes=creds_json.get('scopes'),
            expiry=datetime.datetime.fromtimestamp(
                creds_json.get('token_expiry')
            ),
        )
      except (IOError, TypeError, ValueError, KeyError):
        pass
    return creds

  def _write_creds_to_file(self):
    """Write credentials to file."""
    if self._creds:
      expiry = self._creds.expiry
      if not expiry:
        expiry = datetime.datetime.now()
      creds_json = {
          'access_token': self._creds.token,
          'refresh_token': self._creds.refresh_token,
          'id_token': self._creds.id_token,
          'token_uri': self._creds.token_uri,
          'client_id': self._creds.client_id,
          'client_secret': self._creds.client_secret,
          'scopes': self._creds.scopes,
          'token_expiry': expiry.timestamp(),
      }
      try:
        with open(self._saved_creds_file, 'w') as f:
          json.dump(creds_json, f)
      except (IOError, TypeError, ValueError, RecursionError):
        pass



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Drive CSE API."""

import email.encoders
import email.generator
import email.mime.application
import email.mime.multipart
import email.mime.nonmultipart
import io
import json
import urllib.parse
import google
import google.oauth2.service_account
import google_auth_httplib2
from googleapiclient import http
from googleapiclient import model
import httplib2


class CseDriveClient(object):
  """Drive CSE API Client."""

  GOOGLEAPIS_URL = 'https://www.googleapis.com'

  DRIVE_FILES = 'drive/v3beta/files'
  FILES_METADATA_URL = f'{GOOGLEAPIS_URL}/{DRIVE_FILES}'
  FILES_UPLOAD_URL = f'{GOOGLEAPIS_URL}/upload/{DRIVE_FILES}'

  ENCRYPTED_MIME_TYPE_PREFIX = 'application/vnd.google-gsuite.encrypted'
  OCTET_STREAM = 'application/octet-stream'

  HEADERS = {
      'accept': 'application/json',
      'accept-encoding': 'gzip, deflate',
      'user-agent': '(gzip)',
  }
  PARAMS = {'alt': 'json', 'supportsTeamDrives': 'true'}

  SCOPES = ['https://www.googleapis.com/auth/drive']

  def __init__(self, sa_key_file):
    """Constructor.

    Args:
        sa_key_file: Service-Account Key File path.
    """
    self._creds = (
        google.oauth2.service_account.Credentials.from_service_account_file(
            sa_key_file, scopes=self.SCOPES
        )
    )
    self._model = model.JsonModel()
    self._http = httplib2.Http()
    self._delegated_user_email = None

  def set_delegated_user(self, delegated_user_email):
    """Set the delegate user for future calls to Drive.

    Args:
        delegated_user_email: The delegated user. All subsequent calls to this
          CseDriveClient, until this method is called again, will be on behalf
          of this user. Must be called before making calls to Drive.
    """
    self._delegated_user_email = delegated_user_email

  def generate_cse_token(self, parent_id=None):
    """Generate a CSE Token.

    Args:
        parent_id: Generate a token for a file in this parent in Drive. If None,
          use the root MyDrive of `delegated_user_email`.

    Returns:
        A CSE Token, as a dict.
    """
    if not self._delegated_user_email:
      raise ValueError('Delegated user email not set')
    creds = self._creds.with_subject(self._delegated_user_email)
    authorized_http = google_auth_httplib2.AuthorizedHttp(
        creds, http=self._http
    )

    params = self._new_params({'role': 'writer'})
    if parent_id:
      params['parentId'] = parent_id
    url = f'{self.FILES_METADATA_URL}/generateCseToken?{urllib.parse.urlencode(params)}'
    headers = self._new_headers({})

    request = http.HttpRequest(
        http=authorized_http,
        postproc=self._model.response,
        uri=url,
        method='GET',
        body=None,
        headers=headers,
        methodId='drive.files.generateCseToken',
    )

    response = request.execute()
    return response

  def cse_upload(
      self,
      data_fp,
      file_id,
      filename,
      encryption_details,
      parent_id,
      content_type,
  ):
    """Upload the given data as a new CSE file.

    Args:
      data_fp: The ciphertext data to upload; a readable file-like object.
      file_id: The Drive id of the file being uploaded.
      filename: The name of the file being uploaded.
      encryption_details: Encryption metadata for the file being uploaded.
      parent_id: The parent of the file being uploaded. If None, use the root
        MyDrive of `delegated_user_email`.
      content_type: The content-type of the file being uploaded.

    Returns:
      The metadata of the newly uploaded file, as a dict.
    """
    if not self._delegated_user_email:
      raise ValueError('Delegated user email not set')
    creds = self._creds.with_subject(self._delegated_user_email)
    authorized_http = google_auth_httplib2.AuthorizedHttp(
        creds, http=self._http
    )

    params = self._new_params({'uploadType': 'multipart'})
    url = f'{self.FILES_UPLOAD_URL}?{urllib.parse.urlencode(params)}'

    file_metadata = {
        'id': file_id,
        'name': filename,
        'filename': filename,
        'mimeType': self._get_cse_mime_type(content_type),
        'clientEncryptionDetails': encryption_details,
    }
    if parent_id:
      file_metadata['parents'] = [parent_id]

    msg_root = self._build_media_upload_message(file_metadata, data_fp)
    body = self._get_message_body(msg_root)
    boundary = msg_root.get_boundary()

    content_type = f'multipart/related; boundary="{boundary}"'
    headers = self._new_headers({'content-type': content_type})

    request = http.HttpRequest(
        http=authorized_http,
        postproc=self._model.response,
        uri=url,
        method='POST',
        body=body,
        headers=headers,
        methodId='drive.files.insert',
    )

    response = request.execute()
    return response

  def cse_download(self, file_id):
    """Download the (encrypted) content and metadata of a CSE file.

    Args:
      file_id: The Drive id of the file to download.

    Returns:
      The metadata of the downloaded file, as a dict, and the file's content,
      as a BytesIO object. The caller must close() the object.
    """
    creds = self._creds.with_subject(self._delegated_user_email)
    authorized_http = google_auth_httplib2.AuthorizedHttp(
        creds, http=self._http
    )
    fields = [
        'id',
        'name',
        'mimeType',
        'clientEncryptionDetails/*',
    ]
    params = self._new_params({'fields': ','.join(fields)})
    url = (
        f'{self.FILES_METADATA_URL}/{file_id}?{urllib.parse.urlencode(params)}'
    )
    headers = self._new_headers({})

    request = http.HttpRequest(
        http=authorized_http,
        postproc=self._model.response,
        uri=url,
        method='GET',
        body=None,
        headers=headers,
        methodId='drive.files.get',
    )
    response = request.execute()
    metadata = response

    params = self._new_params({'alt': 'media'})
    url = (
        f'{self.FILES_METADATA_URL}/{file_id}?{urllib.parse.urlencode(params)}'
    )
    request = http.HttpRequest(
        http=authorized_http,
        postproc=None,
        uri=url,
        method='GET',
        body=None,
        headers=headers,
        methodId='drive.files.get',
    )
    buf = io.BytesIO()
    downloader = http.MediaIoBaseDownload(
        buf, request, chunksize=http.DEFAULT_CHUNK_SIZE
    )
    done = False
    while not done:
      _, done = downloader.next_chunk()
    return (metadata, buf)

  def delete(self, file_id):
    """Delete a file.

    Args:
      file_id: The Drive id of the file to delete.

    Returns:
      Empty dict.
    """
    creds = self._creds.with_subject(self._delegated_user_email)
    authorized_http = google_auth_httplib2.AuthorizedHttp(
        creds, http=self._http
    )
    params = self._new_params({})
    url = (
        f'{self.FILES_METADATA_URL}/{file_id}?{urllib.parse.urlencode(params)}'
    )
    headers = self._new_headers({})

    request = http.HttpRequest(
        http=authorized_http,
        postproc=self._model.response,
        uri=url,
        method='DELETE',
        body=None,
        headers=headers,
        methodId='drive.files.delete',
    )
    response = request.execute()
    return response

  def get_jwt_payload(self, jwt):
    """Extract the payload from a JWT.

    Args:
      jwt: A JSON Web Token.

    Returns:
      The payload part of a given jwt, as a dict.
    """
    return google.auth.jwt.decode(token=jwt, verify=False)

  def new_encryption_details(self, kacls_id, wdek, resource_key_hash):
    """Create new encryption-details metadata for a CSE file.

    Args:
      kacls_id: The Drive ID of the KACLS to use.
      wdek: The wrapped data encryption key.
      resource_key_hash: A digest of the encryption key and the encrypted
        resource.

    Returns:
      Encryption-details metadata, as a dict.
    """
    decryption_metadata = {
        'wrappedKey': wdek,
        'kaclsId': kacls_id,
        'aes256GcmChunkSize': 'default',
        'keyFormat': 'tinkAesGcmKey',
        'encryptionResourceKeyHash': resource_key_hash,
    }
    return {
        'encryptionState': 'encrypted',
        'decryptionMetadata': decryption_metadata,
    }

  def _new_params(self, more_params):
    params = self.PARAMS.copy()
    params.update(more_params)
    return params

  def _new_headers(self, more_headers):
    headers = self.HEADERS.copy()
    headers.update(more_headers)
    return headers

  def _get_cse_mime_type(self, content_type=None):
    if not content_type:
      content_type = self.OCTET_STREAM
    return f'{self.ENCRYPTED_MIME_TYPE_PREFIX}; content="{content_type}"'

  def _build_media_upload_message(self, file_metadata, data_fp):
    """Build the media upload message."""
    # multipart/related upload
    msg_root = email.mime.multipart.MIMEMultipart('related')
    setattr(msg_root, '_write_headers', lambda self: None)

    # 1st part: metadata
    msg = email.mime.nonmultipart.MIMENonMultipart('application', 'json')
    msg.set_payload(json.dumps(file_metadata))
    msg_root.attach(msg)

    # 2nd part: data
    media_upload = http.MediaIoBaseUpload(
        fd=data_fp, mimetype=self.OCTET_STREAM
    )
    payload = media_upload.getbytes(0, media_upload.size())
    msg = email.mime.application.MIMEApplication(
        _data=payload, _encoder=email.encoders.encode_noop
    )
    msg['Content-Transfer-Encoding'] = 'binary'
    filename = file_metadata['filename']
    msg['Content-Disposition'] = (
        f'form-data; name="upload"; filename="{filename}"'
    )
    msg_root.attach(msg)

    return msg_root

  def _get_message_body(self, msg_root):
    # encode the body: note that we can't use `as_string`, because
    # it plays games with `From ` lines.
    fp = io.BytesIO()
    g = _BytesGenerator(fp, mangle_from_=False)
    g.flatten(msg_root, unixfrom=False)
    body = fp.getvalue()
    fp.close()
    return body


class _BytesGenerator(email.generator.BytesGenerator):
  _write_lines = email.generator.BytesGenerator.write



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from drive_cse_upload import _cse_kacls_client
from absl.testing import absltest


class CseKaclsClientTest(absltest.TestCase):

  def test_resource_key_hash(self):
    client = _cse_kacls_client.CseKaclsClient()
    key = b'data-encryption-key'
    resource_name = 'resource-name'
    digest = 'BzUyXkZ0424Ve8NtOB1CWgg5hVx60OAPf9SNbVPdtUo='
    self.assertEqual(client.resource_key_hash(key, resource_name), digest)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CSE Tink Keyset utils."""

import tink
from tink import aead
from tink import cleartext_keyset_handle
from tink.proto import tink_pb2

def create_keyset_handle_from_aes_gcm_key(key):
  aes_gcm_key_type_url = aead.aead_key_templates.AES256_GCM_RAW.type_url
  key_id = 1;
  keyset = tink_pb2.Keyset()
  keyset.primary_key_id = key_id
  pkey = keyset.key.add()
  pkey.status = tink_pb2.KeyStatusType.ENABLED
  pkey.output_prefix_type = tink_pb2.OutputPrefixType.RAW
  pkey.key_id = key_id
  pkey.key_data.type_url = aes_gcm_key_type_url
  pkey.key_data.value = key
  pkey.key_data.key_material_type = tink_pb2.KeyData.KeyMaterialType.SYMMETRIC
  return cleartext_keyset_handle.from_keyset(keyset)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CSE Data Encrypter/Decrypter."""

import enum
from drive_cse_upload import _cse_keyset_util
import tink
from tink import aead


class CseCrypter(object):
  """CSE Data Encrypter/Decrypter."""

  SAFE_CHUNK_COUNT_LIMIT = 1 << 30
  AEAD_OVERHEAD = 28
  PLAINTEXT_CHUNK_SIZE = 2097152
  CIPHERTEXT_CHUNK_SIZE = PLAINTEXT_CHUNK_SIZE + AEAD_OVERHEAD
  MAGIC_HEADER = bytes([0x99, 0x5E, 0xCC, 0x5E])

  class OpType(enum.Enum):
    ENCRYPT = 1
    DECRYPT = 2

  def __init__(self):
    self._key = None
    self._aead: aead.Aead = None
    self._current_chunk_count = 0
    aead.register()

  def encrypt(self, plaintext, ciphertext):
    """Encrypt plaintext data.

    Args:
      plaintext: Data to encrypt; a readable buffered file-like object.
      ciphertext: Encrypted data; a writable buffered file-like object.

    Returns:
        None
    """
    key_template = aead.aead_key_templates.AES256_GCM_RAW
    keyset_handle = tink.new_keyset_handle(key_template)
    self._key = self._get_key_from_handle(keyset_handle)
    self._aead = keyset_handle.primitive(aead.Aead)
    self._current_chunk_count = 0

    ciphertext.write(self.MAGIC_HEADER)
    buf = bytearray(self.PLAINTEXT_CHUNK_SIZE)
    while True:
      n = plaintext.readinto(buf)
      data = self._prefix(buf, n)
      # pylint: disable=g-explicit-length-test
      if n < self.PLAINTEXT_CHUNK_SIZE or len(plaintext.peek(1)) == 0:
        ciphertext.write(
            self._update(data, self.OpType.ENCRYPT, is_last_chunk=True)
        )
        break
      ciphertext.write(
          self._update(data, self.OpType.ENCRYPT, is_last_chunk=False)
      )

  def decrypt(self, key, ciphertext, plaintext):
    """Decrypt ciphertext data.

    Args:
      key: Key used to encrypt the ciphertext.
      ciphertext: Data to decrypt; a readable buffered file-like object.
      plaintext: Decrypted data; a writable buffered file-like object.

    Returns:
        None
    """
    keyset_handle = _cse_keyset_util.create_keyset_handle_from_aes_gcm_key(key)
    self._aead = keyset_handle.primitive(aead.Aead)
    self._current_chunk_count = 0

    magic = ciphertext.read(len(self.MAGIC_HEADER))
    if len(magic) != len(self.MAGIC_HEADER) or magic != self.MAGIC_HEADER:
      raise ValueError('Bad input')
    buf = bytearray(self.CIPHERTEXT_CHUNK_SIZE)
    while True:
      n = ciphertext.readinto(buf)
      data = self._prefix(buf, n)
      # pylint: disable=g-explicit-length-test
      if n < self.CIPHERTEXT_CHUNK_SIZE or len(ciphertext.peek(1)) == 0:
        plaintext.write(
            self._update(data, self.OpType.DECRYPT, is_last_chunk=True)
        )
        break
      plaintext.write(
          self._update(data, self.OpType.DECRYPT, is_last_chunk=False)
      )

  def get_key(self):
    """Get the encryption key.

    Returns:
      The key used in last `encrypt` call.
    """
    return self._key

  def _update(self, data, optype, is_last_chunk):
    """Encrypt/Decrypt the next chunk."""
    if self._current_chunk_count + 1 > self.SAFE_CHUNK_COUNT_LIMIT:
      raise ValueError('Input too large')
    associated_data = self._get_associated_data(
        self._current_chunk_count, is_last_chunk
    )
    match optype:
      case self.OpType.ENCRYPT:
        chunk = self._aead.encrypt(data, associated_data)
      case self.OpType.DECRYPT:
        chunk = self._aead.decrypt(data, associated_data)
    self._current_chunk_count += 1
    return chunk

  def _get_key_from_handle(self, keyset_handle):
    keyset = keyset_handle._keyset  # pylint: disable=protected-access
    key = keyset.key[0]
    key_data = key.key_data
    return key_data.value

  def _prefix(self, buf, n):
    assert n <= len(buf)
    if n == len(buf):
      return bytes(buf)
    return bytes(buf[0:n])

  def _get_associated_data(self, chunk_index, is_last_chunk):
    data = bytearray(chunk_index.to_bytes(4, 'big'))
    if is_last_chunk:
      data.append(1)
    else:
      data.append(0)
    return bytes(data)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CSE Uploader."""

import hashlib
import io
import mimetypes
import os
import pathlib
from typing import Any
from drive_cse_upload import _cse_crypter
from drive_cse_upload import _cse_drive_client
from drive_cse_upload import _cse_idp_client
from drive_cse_upload import _cse_kacls_client
from drive_cse_upload import _cse_util


class CseUploader(object):
  """Uploader of CSE files to Drive."""

  def __init__(
      self,
      sa_key_file: str,
      client_secret_file: str,
      saved_creds_file: str,
      validate: bool | None = True,
  ):
    """Constructor.

    Note: Arguments `sa_key_file`, `client_secret_file`, and `saved_creds_file`
    denote files containing sensitive information that should be protected.
    Users must ensure that the files passed-in / created here are not readable
    by anyone but their owner.

    Args:
        sa_key_file: Service-Account Private Key File path; callers must
          download the file from the Google Cloud Console and store it to this
          path.
        client_secret_file: Client Secret File path for the configured IDP OAuth
          Client-Id; callers must download the file from their IDP and store it
          to this path.
        saved_creds_file: Where to store the IDP Oauth credentials, created by
          this package.
        validate: If True (or omitted), validate every uploaded file by
          downloading and decrypting it. If validation of an uploaded file
          fails, delete it.
    """
    self._sa_key_file = os.path.abspath(sa_key_file)
    self._idp = _cse_idp_client.CseIdpClient(
        os.path.abspath(client_secret_file), os.path.abspath(saved_creds_file)
    )
    self._validate = validate

  def upload(
      self,
      input_file: str,
      delegated_user_email: str,
      parent_id: str | None = None,
  ) -> dict[str, Any]:
    """Upload a single file from local-host as a CSE file in Drive.

    Args:
        input_file: Local input file to upload.
        delegated_user_email: Upload the file as this user in Drive.
        parent_id: If not None, upload the file as a child of this parent
          (folder or shared-drive) in Drive. If omitted or None, upload the file
          to the root MyDrive of `delegated_user_email`.

    Returns:
      The metadata of the newly uploaded file, as a dict.

    Raises:
      RuntimeError: The uploaded file validation failed.
    """

    crypter = _cse_crypter.CseCrypter()
    kacls = _cse_kacls_client.CseKaclsClient()
    driver = _cse_drive_client.CseDriveClient(self._sa_key_file)

    content_type = mimetypes.guess_type(input_file)[0]
    filename = pathlib.Path(input_file).name

    driver.set_delegated_user(delegated_user_email)

    cse_token = driver.generate_cse_token(parent_id)
    jwt = cse_token['jwt']
    file_id = cse_token['fileId']
    kacls_id = cse_token['currentKaclsId']
    jwt_payload = driver.get_jwt_payload(jwt)
    resource_name = jwt_payload['resource_name']
    kacls_url = jwt_payload['kacls_url']
    perimeter_id = jwt_payload.get('perimeter_id', '')

    authn = self._idp.get_authn_token()

    outp = io.BytesIO()
    input_plaintext_digest = None
    with open(input_file, 'rb') as inp:
      crypter.encrypt(inp, outp)
      if self._validate:
        inp.seek(0)
        input_plaintext_digest = hashlib.file_digest(
            inp, hashlib.sha256
        ).hexdigest()  # pytype: disable=wrong-arg-types
    key = crypter.get_key()
    wdek = kacls.privileged_wrap(
        _cse_util.b64encode(key), resource_name, authn, kacls_url, perimeter_id
    )
    resource_key_hash = kacls.resource_key_hash(key, resource_name)

    encryption_details = driver.new_encryption_details(
        kacls_id,
        _cse_util.b64urlencode(wdek.encode()),
        _cse_util.b64urlencode(resource_key_hash.encode()),
    )
    decryption_metadata = encryption_details['decryptionMetadata']

    file = driver.cse_upload(
        outp, file_id, filename, encryption_details, parent_id, content_type
    )
    outp.close()

    if self._validate:
      drive_id = file['id']
      metadata, content = driver.cse_download(drive_id)

      downloaded_encryption_details = metadata['clientEncryptionDetails']
      downloaded_decryption_metadata = downloaded_encryption_details[
          'decryptionMetadata'
      ]
      downloaded_jwt = downloaded_decryption_metadata['jwt']
      downloaded_wdek = _cse_util.b64urldecode(
          downloaded_decryption_metadata['wrappedKey']
      ).decode()
      downloaded_jwt_payload = driver.get_jwt_payload(downloaded_jwt)
      downloaded_resource_name = downloaded_jwt_payload['resource_name']
      downloaded_kacls_url = downloaded_jwt_payload['kacls_url']
      authn = self._idp.get_authn_token()
      downloaded_key = kacls.privileged_unwrap(
          downloaded_wdek,
          downloaded_resource_name,
          authn,
          downloaded_kacls_url,
      )

      outp = io.BytesIO()
      content.seek(0)
      inp = io.BufferedReader(content)  # pytype: disable=wrong-arg-types
      crypter.decrypt(_cse_util.b64decode(downloaded_key), inp, outp)
      outp.seek(0)
      downloaded_plaintext_digest = hashlib.file_digest(
          outp, hashlib.sha256
      ).hexdigest()
      content.close()
      inp.close()
      outp.close()

      for k in list(downloaded_decryption_metadata.keys()):
        if k not in decryption_metadata:
          del downloaded_decryption_metadata[k]
      if downloaded_encryption_details != encryption_details:
        driver.delete(drive_id)
        raise RuntimeError('Uploaded file metadata validation failed')
      if downloaded_plaintext_digest != input_plaintext_digest:
        driver.delete(drive_id)
        raise RuntimeError('Uploaded file decryption failed')

    return file



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import random
from drive_cse_upload import _cse_crypter
from absl.testing import absltest


class CseCrypterTest(absltest.TestCase):

  def test_encrypt_short(self):
    # single chunk
    data = b'Hello, CSE World!'
    self._do_test_encrypt(data)

  def test_encrypt_long(self):
    chunk_size = _cse_crypter.CseCrypter.PLAINTEXT_CHUNK_SIZE
    # at least 2 chunks
    length = random.randint(2 * chunk_size, 5 * chunk_size)
    data = random.randbytes(length)
    self._do_test_encrypt(data)

  def test_encrypt_empty(self):
    # 0 size
    data = b''
    self._do_test_encrypt(data)

  def test_encrypt_chunk_size(self):
    # exact positive multiple of chunk size
    chunk_size = _cse_crypter.CseCrypter.PLAINTEXT_CHUNK_SIZE
    length = random.randint(1, 3) * chunk_size
    data = random.randbytes(length)
    self._do_test_encrypt(data)

  def _do_test_encrypt(self, data):
    crypter = _cse_crypter.CseCrypter()
    plaintext = io.BufferedReader(io.BytesIO(data))
    ciphertext = io.BytesIO()
    crypter.encrypt(plaintext, ciphertext)
    plaintext.seek(0)
    ciphertext.seek(0)
    plaintext2 = io.BytesIO()
    crypter.decrypt(
        crypter.get_key(), io.BufferedReader(ciphertext), plaintext2
    )
    plaintext2.seek(0)
    self.assertEqual(plaintext2.read(), plaintext.read())
    plaintext.close()
    ciphertext.close()
    plaintext2.close()

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CSE Utilities."""

import base64


def b64encode(data: bytes) -> str:
  """B64 encode.

  Args:
      data: Bytes to encode.

  Returns:
      A base64-encoded string of the data bytes.
  """
  return base64.b64encode(data).decode()


def b64urlencode(data: bytes) -> str:
  """B64 URL encode.

  Args:
      data: Bytes to encode.

  Returns:
      A URL-base64-encoded string of the data bytes.
  """
  return base64.b64encode(data, altchars=b'-_').decode()


def b64decode(data: str) -> bytes:
  """B64 decode.

  Args:
      data: String to decode.

  Returns:
      A base64-decoded bytes of the data string.
  """
  return base64.b64decode(data)


def b64urldecode(data: str) -> bytes:
  """B64 URL decode.

  Args:
      data: String to decode.

  Returns:
      A URL-base64-decoded bytes of the data string.
  """
  return base64.b64decode(data, altchars=b'-_')
